@@ -54,7 +54,20 @@ class PytestMergify:
             )
             return
 
-        self.mergify_ci.report_flaky_detection(terminalreporter)
+        if self.mergify_ci.flaky_detector:
+            terminalreporter.write_line(self.mergify_ci.flaky_detector.make_report())
+        elif self.mergify_ci.flaky_detector_error_message:
+            terminalreporter.write_line(
+                f"""⚠️ Flaky detection couldn't be enabled because of an error.
+
+Common issues:
+  • Your 'MERGIFY_TOKEN' might not be set or could be invalid
+  • There might be a network connectivity issue with the Mergify API
+
+📚 Documentation: https://docs.mergify.com/ci-insights/test-frameworks/pytest/
+🔍 Details: {self.mergify_ci.flaky_detector_error_message}""",
+                yellow=True,
+            )
 
         # CI Insights Quarantine warning logs
         if not self.mergify_ci.branch_name:
@@ -123,6 +136,13 @@ class PytestMergify:
             )
         self.has_error = False
 
+        if self.mergify_ci.flaky_detector:
+            self.mergify_ci.flaky_detector.set_deadline()
+
+    def pytest_collection_finish(self, session: _pytest.main.Session) -> None:
+        if self.mergify_ci.flaky_detector:
+            self.mergify_ci.flaky_detector.filter_existing_tests_with_session(session)
+
     @pytest.hookimpl(hookwrapper=True)
     def pytest_sessionfinish(
         self,
@@ -132,9 +152,6 @@ class PytestMergify:
             yield
             return
 
-        # Execute flaky detection just before ending the session.
-        self.mergify_ci.run_flaky_detection(session)
-
         yield
 
         self.session_span.set_status(
@@ -144,51 +161,76 @@ class PytestMergify:
         )
         self.session_span.end()
 
-    def _attributes_from_item(
+    def _get_item_attributes(
         self, item: _pytest.nodes.Item
-    ) -> typing.Dict[str, typing.Union[str, int]]:
+    ) -> typing.Dict[str, typing.Any]:
         filepath, line_number, testname = item.location
         namespace = testname.replace(item.name, "")
         if namespace.endswith("."):
             namespace = namespace[:-1]
 
-        return {
+        result = {
             SpanAttributes.CODE_FILEPATH: filepath,
             SpanAttributes.CODE_FUNCTION: item.name,
             SpanAttributes.CODE_LINENO: line_number or 0,
             SpanAttributes.CODE_NAMESPACE: namespace,
             "code.file.path": str(_pytest.pathlib.absolutepath(item.reportinfo()[0])),
             "code.line.number": line_number or 0,
+            "test.scope": "case",
         }
 
+        if _should_skip_item(item):
+            result["cicd.test.quarantined"] = False
+            result["test.case.result.status"] = "skipped"
+        else:
+            result["cicd.test.quarantined"] = (
+                self.mergify_ci.mark_test_as_quarantined_if_needed(item)
+            )
+
+        return result
+
     @pytest.hookimpl(hookwrapper=True)
     def pytest_runtest_protocol(
-        self, item: _pytest.nodes.Item
+        self,
+        item: _pytest.nodes.Item,
     ) -> typing.Generator[None, None, None]:
         if not self.tracer:
             yield
             return
 
-        additional_attributes: typing.Dict[str, typing.Any] = {}
-        if _should_skip_item(item):
-            additional_attributes["test.case.result.status"] = "skipped"
-            additional_attributes["cicd.test.quarantined"] = False
-        else:
-            quarantined = self.mergify_ci.mark_test_as_quarantined_if_needed(item)
-            additional_attributes["cicd.test.quarantined"] = quarantined
-
-        context = opentelemetry.trace.set_span_in_context(self.session_span)
         with self.tracer.start_as_current_span(
             item.nodeid,
-            attributes={
-                **self._attributes_from_item(item),
-                **additional_attributes,
-                **{"test.scope": "case"},
-            },
-            context=context,
+            attributes=self._get_item_attributes(item),
+            context=opentelemetry.trace.set_span_in_context(self.session_span),
         ):
             yield
 
+    @pytest.hookimpl(tryfirst=True)
+    def pytest_runtest_teardown(self, item: _pytest.nodes.Item) -> None:
+        # NOTE(remyduthu):
+        # We run flaky detection during the teardown phase because setup has
+        # already finished and we only need to focus on the call phase. At this
+        # point, we know how long the initial execution took, which lets us
+        # calculate the number of retries. This timing also ensures all
+        # fixtures, including session-scoped ones, are still available before
+        # the item is completely torn down.
+        if not self.tracer or not self.mergify_ci.flaky_detector:
+            return
+
+        attributes = self._get_item_attributes(item)
+        context = opentelemetry.trace.set_span_in_context(self.session_span)
+
+        for _ in range(
+            self.mergify_ci.flaky_detector.get_retry_count_for_new_test(item.nodeid)
+        ):
+            if self.mergify_ci.flaky_detector.is_deadline_exceeded():
+                return
+
+            with self.tracer.start_as_current_span(
+                item.nodeid, attributes=attributes, context=context
+            ):
+                _pytest.runner.call_and_report(item=item, log=True, when="call")
+
     def pytest_exception_interact(
         self,
         node: _pytest.nodes.Node,
@@ -243,7 +285,10 @@ class PytestMergify:
             }
         )
 
-        self.mergify_ci.handle_flaky_detection_for_report(report)
+        if self.mergify_ci.flaky_detector:
+            detected = self.mergify_ci.flaky_detector.detect_from_report(report)
+            if detected:
+                test_span.set_attributes({"cicd.test.new": True})
 
 
 def pytest_addoption(parser: _pytest.config.argparsing.Parser) -> None:
